/***************** EMSEXEC.C ******************
*  Program to execute code in Expanded Memory
*
*  Must be linked to EMS.C (from Part 1)
*
*  Validated for Turbo C --- use large model
*
*  Author: Andrew Binstock (c) 1990 C Gazette
*  Use freely but acknowledge author and publication.
**********************************************/

#include <stdio.h>
#include <stdlib.h>
#include <dos.h>

#include "ems.h"        /* Prototypes from Part 1 */

#ifdef MSC              /* If Microsoft, make sure of */
#pragma pack (1);       /* NO structure padding.      */
#endif

#define FAR far        /* For portability */

#define MAP    0x0000  /* For service 50h of EMS interrupt */
#define UNMAP  0xFFFF  /* which maps and unmaps pages      */

union REGS inregs, outregs;
struct SREGS segs;

extern int Handle;     /* in EMS.C -- see article  */

char FAR * frame_ptr;        /* ptr to EMM frame */

int pages_needed = 1;

struct page_set {
  unsigned short logical;
  unsigned short physical;
};

unsigned int frame_segment;

struct page_set pages [ 4 ];
struct page_set map_call_pages [4];

int  EmsAlterPageMapAndCall ( struct page_set *, char,
                              struct page_set *, char);
int  EmsGetVersion     ( void );
void EmsMapUnmapPages  ( struct page_set *, int, int, unsigned int );
void load_overlay      ( char *, char **, unsigned int );

#define DATA_SIZE   15
#define CODE_SIZE   13

/* The following arrays contain the data and the code which we
   will load and execute in expanded memory. The code portion
   is the actual hex codes generated by an assembler. The code
   simply prints the string in data_array and does a far return.
   Notice that since we are using the DOS print-string service,
   our string ends in a '$'. The character-by-character initialization
   of data_array means that our "string" has no terminating null byte.
 */

char data_array [ DATA_SIZE ]   =
    {'H','e','l','l','o',' ','f','r','o','m',' ','E','M','S','$'};

char code_array [ CODE_SIZE ]   =
    {
      0xB8, 0x00, 0xD0,    /* mov ax, D000h ; load address into DS:DX */
                           /* see below for changes to this value     */

      0x8E, 0xD8,          /* mov ds, ax                              */
      0xBA, 0x00, 0x00,    /* mov dx, 00h   ; the offset              */
      0xB4, 0x09,          /* mov ah, 09    ; print string service    */
      0xCD, 0x21,          /* int 21h                                 */
      0xCB                 /* retf          ; far return              */
    };

int main ( void )
{
    int i, j;

    if ( !EmsInstalled() )        /* First check for EMM */
    {
        fprintf ( stderr, "EMS not found\n" );
        return ( 5 );
    }

    if ( EmsGetVersion() < 400 )
    {
        fprintf ( stderr, "Requires EMM version 4.0 or later\n" );
        return ( 6 );
    }

    if (( i = EmsPagesAvail()) < pages_needed )
    {
        fprintf ( stderr, "Need %d pages, Found only %d\n",
                    pages_needed, i );
        exit ( 7 );
    }

    if (( frame_ptr = EmsAlloc ( pages_needed )) == NULL )
    {
        fprintf ( stderr, "Error allocating EMS pages\n" );
        exit ( 8 );
    }

    EmsMapUnmapPages ( pages, Handle, 1, MAP ); /* Map in one page */

    /* Load the frame segment into the code array. Bytes 2 and 3
       of the code array specify the segment portion of the data
       segment address. In our example, since we load data into
       the first bytes of the page frame (see the for loops after
       the next comment block), the data segment address is the
       page frame address.

       Note that because of Intel byte-order scheme, the low-order
       byte of the segment address is loaded first, then the high-order
       byte.
     */

    frame_segment  = FP_SEG( frame_ptr );
    code_array [1] = (char) frame_segment;
    code_array [2] = (char) ( frame_segment >> 8 );

    /* Load executable: first data, then code right behind it. */

    for ( i = 0; i < DATA_SIZE; i++ )
        frame_ptr[i] = data_array[i];

    for ( j = 0, i = DATA_SIZE; j < CODE_SIZE; i++, j++ )
        frame_ptr[i] = code_array[j];

   /* Specify which logical and physical page is to be mapped
      after the call and replaced (remapped) after the return.
      We are only using page 0, so both are set to 0.
    */

    map_call_pages[0].logical  = 0;
    map_call_pages[0].physical = 0;

    /* Transfer control to the loaded module; i.e., call it.
       Control returns after the RETF instrction is hit.
       RETF = far return. This means that any loaded code
       will require a far return, i.e., some portion will
       have to be in assembly language.
     */

    i = EmsAlterPageMapAndCall ( map_call_pages, 1, pages, 1 );

    /* Check for any errors. Errors use EMM's error scheme (see
       article): 0 = all OK, 0x80 - 0x8F error detected by EMM
     */

    if ( i != 0)
        fprintf ( stderr, "\nError on call to loaded code: %x\n", i );

    EmsFree ();   /* Deallocate pages we used */

    return ( EXIT_SUCCESS );
}

int EmsGetVersion ( void )  /* Return the EMM version as version * 100 */
{
    inregs.h.ah = 0x46;
    int86 ( 0x67, &inregs, &outregs );  /* call EMM */

    if ( outregs.h.ah != 0 ) /* Did an error occur? */
    {
        fprintf ( stderr, "Error %x accessing EMM\n", outregs.h.ah );
        return ( 7 );
    }

    /* Major version in upper 4 bits, minor version in lower four. */
    return (( outregs.h.al >> 4 ) * 100 +
            ( outregs.h.al & 0x00FF ));
}

void EmsMapUnmapPages  ( struct page_set *page,  /* page set to map     */
                         int handle,             /* handle to map to    */
                         int how_many,           /* # of pages to map   */
                         unsigned int action)    /* map or unmap ?      */
{
    int i;
    struct page_set *orig_page = page;      /* save initial value  */

    for ( i = 0 ; i < how_many ; i++ )      /* initialize the page */
    {                                       /* info structure.     */
        page->physical = i;
        page->logical  = ( action == MAP ? i : UNMAP );

        page += 1;                          /* onto the next page  */
    }

    inregs.x.ax = 0x5000;                   /* the EMM map service */
    inregs.x.dx = handle;
    inregs.x.cx = how_many;
    inregs.x.si = FP_OFF(orig_page);
    segs.ds     = FP_SEG(orig_page);
    int86x ( 0x67, &inregs, &outregs, &segs );

    if ( outregs.h.ah != 0 )
    {
        fprintf ( stderr, "Page mapping Failed: %x\n", outregs.h.ah );
        exit ( 9 );
    }
}

int EmsAlterPageMapAndCall ( struct page_set *new_page_map_ptr,
                             char             new_page_map_length,
                             struct page_set *old_page_map_ptr,
                             char             old_page_map_length )
{
    struct map_and_call_struct {
        char FAR            * target_ptr; /* where control is transferred */

        char                  new_page_map_len;
        struct page_set FAR * new_page_map;

        char                  old_page_map_len;
        struct page_set FAR * old_page_map;
    }
        map_and_call;

    struct map_and_call_struct *map_and_call_ptr = &map_and_call;

    map_and_call.target_ptr           = (char FAR *) frame_ptr + DATA_SIZE;

    map_and_call.new_page_map_len     = new_page_map_length;
    map_and_call.new_page_map         = new_page_map_ptr;

    map_and_call.old_page_map_len     = old_page_map_length;
    map_and_call.old_page_map         = old_page_map_ptr;

    inregs.h.ah = 0x56;  /* service 56: alter page map and call */
    inregs.h.al = 0;     /* 0 = old/new_page_map_ptr point to page #s,
                            1 = they point to seg:off address of pages */
    inregs.x.dx = Handle;
    segs.ds     = FP_SEG (map_and_call_ptr);  /* DS:SI is map_and_call */
    inregs.x.si = FP_OFF (map_and_call_ptr);

    int86x ( 0x67, &inregs, &outregs, &segs );

    return (( int ) outregs.h.ah );
}